@@ -1,8 +1,11 @@
 import typing
 
-from hyperscan._hyperscan import *
+from hyperscan._ext import *  # noqa: F401, F403
 
-__version__ = "0.3.0"
+try:
+    from hyperscan._version import __version__  # pyright: ignore
+except ImportError:
+    __version__ = "unknown"
 
 
 class ExpressionExt(typing.NamedTuple):
