@@ -7,49 +7,48 @@
 #include <stdio.h>
 #include <structmember.h>
 
-#define ADD_INT_CONSTANT(module, name) \
-  PyModule_AddIntConstant(module, #name, name);
-#define HANDLE_CHIMERA_ERR(err, rv)                \
-  if (err != CH_SUCCESS) {                         \
-    char serr[80];                                 \
-    sprintf(serr, "error code %i", err);           \
-    PyGILState_STATE gstate = PyGILState_Ensure(); \
-    PyErr_SetString(HyperscanError, serr);         \
-    PyGILState_Release(gstate);                    \
-    return rv;                                     \
+#define ADD_INT_CONSTANT(module, name)                    \
+  if (PyModule_AddIntConstant(module, #name, name) < 0) { \
+    Py_XDECREF(module);                                   \
+    return NULL;                                          \
+  };
+#define HANDLE_CHIMERA_ERR(err, rv)                   \
+  if (err != CH_SUCCESS) {                            \
+    char serr[80];                                    \
+    sprintf(serr, "error code %i", err);              \
+    PyGILState_STATE gstate = PyGILState_Ensure();    \
+    PyErr_SetString(HyperscanErrors[abs(err)], serr); \
+    PyGILState_Release(gstate);                       \
+    return rv;                                        \
   }
-#define HANDLE_HYPERSCAN_ERR(err, rv)              \
-  if (err != HS_SUCCESS) {                         \
-    char serr[80];                                 \
-    sprintf(serr, "error code %i", err);           \
-    PyGILState_STATE gstate = PyGILState_Ensure(); \
-    PyErr_SetString(HyperscanError, serr);         \
-    PyGILState_Release(gstate);                    \
-    return rv;                                     \
+#define HANDLE_HYPERSCAN_ERR(err, rv)                 \
+  if (err != HS_SUCCESS) {                            \
+    char serr[80];                                    \
+    sprintf(serr, "error code %i", err);              \
+    PyGILState_STATE gstate = PyGILState_Ensure();    \
+    PyErr_SetString(HyperscanErrors[abs(err)], serr); \
+    PyGILState_Release(gstate);                       \
+    return rv;                                        \
+  }
+#define ADD_HYPERSCAN_ERROR(module, errors, base, name, hs_err, doc) \
+  if (PyModule_AddIntConstant(module, #hs_err, hs_err) < 0) {        \
+    Py_XDECREF(module);                                              \
+    return NULL;                                                     \
+  }                                                                  \
+  PyObject *name =                                                   \
+    PyErr_NewExceptionWithDoc("hyperscan." #name, doc, base, NULL);  \
+  if (name == NULL) {                                                \
+    Py_XDECREF(module);                                              \
+    return NULL;                                                     \
+  } else {                                                           \
+    if (PyModule_AddObject(module, #name, name) < 0) {               \
+      Py_XDECREF(module);                                            \
+      return NULL;                                                   \
+    }                                                                \
+    errors[abs(hs_err)] = name;                                      \
   }
 
-#if PY_MAJOR_VERSION >= 3
-#define MOD_ERROR_VAL NULL
-#define MOD_SUCCESS_VAL(val) val
-#define MOD_INIT(name) PyMODINIT_FUNC PyInit_##name(void)
-#define MOD_DEF(ob, name, doc, methods)   \
-  static struct PyModuleDef moduledef = { \
-    PyModuleDef_HEAD_INIT,                \
-    name,                                 \
-    doc,                                  \
-    -1,                                   \
-    methods,                              \
-  };                                      \
-  ob = PyModule_Create(&moduledef);
-#else
-#define MOD_ERROR_VAL
-#define MOD_SUCCESS_VAL(val)
-#define MOD_INIT(name) void init##name(void)
-#define MOD_DEF(ob, name, doc, methods) ob = Py_InitModule3(name, methods, doc);
-#endif
-
-const char *DEFAULT_VERSION = "unknown";
-
+static PyObject *HyperscanErrors[33] = {NULL};
 static PyObject *HyperscanError;
 static PyTypeObject DatabaseType;
 static PyTypeObject ScratchType;
@@ -246,15 +245,21 @@ static PyObject *Database_compile(
   size_t *lens;
   uint32_t globalflag;
 
-  expressions = PyMem_RawMalloc(elements * sizeof(char *));
+  if (self->chimera && self->ch_db != NULL)
+    ch_free_database(self->ch_db);
+  else if (!self->chimera && self->hs_db != NULL) {
+    hs_free_database(self->hs_db);
+  }
+
+  expressions = malloc(elements * sizeof(char *));
   if (expressions == NULL)
     goto memory_error;
 
-  flags = PyMem_RawMalloc(elements * sizeof(uint32_t));
+  flags = malloc(elements * sizeof(uint32_t));
   if (flags == NULL)
     goto memory_error;
 
-  ids = PyMem_RawMalloc(elements * sizeof(uint32_t));
+  ids = malloc(elements * sizeof(uint32_t));
   if (ids == NULL)
     goto memory_error;
 
@@ -320,7 +325,7 @@ static PyObject *Database_compile(
         "chimera does not support pure literal expressions");
       goto python_error;
     }
-    lens = PyMem_RawMalloc(elements * sizeof(size_t));
+    lens = malloc(elements * sizeof(size_t));
     if (lens == NULL)
       goto memory_error;
     for (uint64_t i = 0; i < elements; i++) {
@@ -338,11 +343,11 @@ static PyObject *Database_compile(
       NULL,
       &self->hs_db,
       &hs_compile_err);
-    PyMem_RawFree(lens);
+    free(lens);
     Py_END_ALLOW_THREADS;
-    PyMem_RawFree(expressions);
-    PyMem_RawFree(flags);
-    PyMem_RawFree(ids);
+    free(expressions);
+    free(flags);
+    free(ids);
 
     if (hs_err != HS_SUCCESS) {
       PyErr_Format(
@@ -369,23 +374,15 @@ static PyObject *Database_compile(
         &self->ch_db,
         &ch_compile_err);
       Py_END_ALLOW_THREADS;
-      PyMem_RawFree(expressions);
-      PyMem_RawFree(flags);
-      PyMem_RawFree(ids);
-      if (ch_err != CH_SUCCESS) {
-        PyErr_Format(
-          HyperscanError,
-          "%s (id:%d)",
-          ch_compile_err->message,
-          hs_compile_err->expression);
-        ch_free_compile_error(ch_compile_err);
-        return NULL;
-      }
+      free(expressions);
+      free(flags);
+      free(ids);
+      HANDLE_CHIMERA_ERR(ch_err, NULL);
     } else {
       if (oext != Py_None) {
-        ext = PyMem_RawMalloc(elements * sizeof(struct hs_expr_ext *));
+        ext = malloc(elements * sizeof(struct hs_expr_ext *));
         for (uint64_t i = 0; i < elements; i++) {
-          ext[i] = PyMem_RawMalloc(sizeof(struct hs_expr_ext));
+          ext[i] = malloc(sizeof(struct hs_expr_ext));
           PyObject *oext_item = PySequence_GetItem(oext, i);
           if (oext_item == NULL) {
             PyErr_Format(
@@ -422,30 +419,43 @@ static PyObject *Database_compile(
         &hs_compile_err);
       Py_END_ALLOW_THREADS;
 
-      PyMem_RawFree(expressions);
-      PyMem_RawFree(flags);
-      PyMem_RawFree(ids);
+      free(expressions);
+      free(flags);
+      free(ids);
       if (hs_err != HS_SUCCESS) {
         PyErr_SetString(HyperscanError, hs_compile_err->message);
         hs_free_compile_error(hs_compile_err);
         return NULL;
       }
     }
-    PyMem_RawFree(ext);
+    free(ext);
   }
 
   if (self->scratch == Py_None) {
     self->scratch =
       PyObject_CallFunction((PyObject *)&ScratchType, "O", (PyObject *)self, 0);
-  } else {
-    Scratch *scratch = ((Scratch *)self->scratch);
-    if (self->chimera) {
-      ch_error_t ch_err = ch_alloc_scratch(self->ch_db, &scratch->ch_scratch);
+  }
+
+  Scratch *scratch = ((Scratch *)self->scratch);
+
+  if (self->chimera) {
+    ch_error_t ch_err;
+    if (scratch->ch_scratch != NULL) {
+      ch_err = ch_free_scratch(scratch->ch_scratch);
       HANDLE_CHIMERA_ERR(ch_err, NULL);
-    } else {
-      hs_error_t hs_err = hs_alloc_scratch(self->hs_db, &scratch->hs_scratch);
-      HANDLE_HYPERSCAN_ERR(hs_err, NULL);
+      scratch->ch_scratch = NULL;
     }
+    ch_err = ch_alloc_scratch(self->ch_db, &scratch->ch_scratch);
+    HANDLE_CHIMERA_ERR(ch_err, NULL);
+  } else {
+    hs_error_t hs_err;
+    if (scratch->hs_scratch != NULL) {
+      hs_err = hs_free_scratch(scratch->hs_scratch);
+      HANDLE_HYPERSCAN_ERR(hs_err, NULL);
+      scratch->hs_scratch = NULL;
+    }
+    hs_err = hs_alloc_scratch(self->hs_db, &scratch->hs_scratch);
+    HANDLE_HYPERSCAN_ERR(hs_err, NULL);
   }
 
   Py_RETURN_NONE;
@@ -454,9 +464,9 @@ memory_error:
   return PyErr_NoMemory();
 
 python_error:
-  PyMem_RawFree(expressions);
-  PyMem_RawFree(flags);
-  PyMem_RawFree(ids);
+  free(expressions);
+  free(flags);
+  free(ids);
   return NULL;
 }
 
@@ -643,13 +653,13 @@ static PyObject *Database_stream(Database *self, PyObject *args, PyObject *kwds)
   PyObject *ocallback = Py_None;
   PyObject *octx = Py_None;
   static char *kwlist[] = {
-    "flags",
     "match_event_handler",
+    "flags",
     "context",
     NULL,
   };
   if (!PyArg_ParseTupleAndKeywords(
-        args, kwds, "|IOO", kwlist, &flags, &ocallback, &octx))
+        args, kwds, "O|IO", kwlist, &ocallback, &flags, &octx))
     return NULL;
   PyObject *stream = PyObject_CallFunction(
     (PyObject *)&StreamType, "OIOO", (PyObject *)self, flags, ocallback, octx);
@@ -660,12 +670,12 @@ static PyObject *Database_stream(Database *self, PyObject *args, PyObject *kwds)
 }
 
 static PyMemberDef Database_members[] = {
-  {"mode", T_INT, offsetof(Database, mode), 0, ":obj:`int`: Scanning mode."},
+  {"mode", T_INT, offsetof(Database, mode), 0, "int: Scanning mode."},
   {"scratch",
    T_OBJECT_EX,
    offsetof(Database, scratch),
    0,
-   ":class:`~.Scratch`: Scratch space object."},
+   ":class:`Scratch`: Scratch space object."},
   {NULL}};
 
 static PyMethodDef Database_methods[] = {
@@ -675,23 +685,22 @@ static PyMethodDef Database_methods[] = {
    "compile(expressions, ids=None, elements=None, flags=0, literal=False)\n\n"
    "    Compiles regular expressions.\n\n"
    "    Args:\n"
-   "        expressions (sequence of :obj:`str`): A sequence of regular\n"
+   "        expressions (sequence of str): A sequence of regular\n"
    "            expressions.\n"
-   "        ids (sequence of :obj:`int`, optional): A sequence of\n"
+   "        ids (sequence of int, optional): A sequence of\n"
    "            expression identifiers.\n"
-   "        elements (:obj:`int`, optional): Length of the expressions\n"
+   "        elements (int, optional): Length of the expressions\n"
    "            sequence.\n"
-   "        flags (sequence of :obj:`int` or :obj:`int`, optional):\n"
+   "        flags (sequence of int or int, optional):\n"
    "            Sequence of flags associated with each expression, or a\n"
    "            single value which is applied to all expressions.\n"
    "        literal (bool, optional): If True, uses the pure literal\n"
    "            expression compiler introduced in Hyperscan 5.2.0.\n\n"
-   "        ext (list, optional): A list of tuples used to define extended\n"
-   "            behavior for each pattern. Tuples must contain **flags**,\n"
-   "            **min_offset**, **max_offset**, **min_length**,\n"
-   "            **edit_distance**, and **hamming_distance**. See hyperscan\n"
-   "            documentation for more information. **Note:** this\n"
-   "            parameter if **literal** is True.\n\n"},
+   "        ext (sequence of tuple, optional): A list of tuples used to "
+   "            define extended behavior for each pattern. Tuples must \n"
+   "            contain **flags**, **min_offset**, **max_offset**, \n"
+   "            **min_length**, **edit_distance**, and **hamming_distance**.\n"
+   "            See hyperscan documentation for more information.\n\n"},
   {"info",
    (PyCFunction)Database_info,
    METH_VARARGS,
@@ -713,30 +722,30 @@ static PyMethodDef Database_methods[] = {
    "scan(data, match_event_handler, flags=0, context=None, scratch=None)\n\n"
    "    Scans a block of text.\n\n"
    "    Args:\n"
-   "        data (:obj:`str`): The block of text to scan, if the database\n"
+   "        data (str): The block of text to scan, if the database\n"
    "            was opened with streaming or block mode, or a list of\n"
    "            buffers (i.e. :obj:`bytearray`) if the database was\n"
    "            opened with vectored mode.\n"
    "        match_event_handler (callable): The match callback, which is\n"
    "            invoked for each match result, and passed the expression\n"
    "            id, start offset, end offset, flags, and a context object.\n"
-   "        flags (:obj:`int`): Currently unused.\n"
+   "        flags (int): Currently unused.\n"
    "        context (:obj:`object`): A context object passed as the last\n"
    "            arg to **match_event_handler**.\n"
-   "        scratch (:class:`~.Scratch`): A scratch object.\n\n"},
+   "        scratch (:class:`Scratch`): A scratch object.\n\n"},
   {"stream",
    (PyCFunction)Database_stream,
    METH_VARARGS | METH_KEYWORDS,
-   "stream(flags=0, match_event_handler=None, context=None)\n\n"
+   "stream(match_event_handler=None, flags=0, context=None)\n\n"
    "    Returns a new stream context manager.\n\n"
    "    Args:\n"
-   "        flags (:obj:`int`): Currently unused.\n"
    "        match_event_handler (callable, optional): The match callback,\n"
    "            which is invoked for each match result, and passed the\n"
    "            expression id, start offset, end offset, flags, and a\n"
    "            context object. Note that this callback will override\n"
    "            the match event handler defined in the\n"
-   "            :class:`~.Database` instance.\n"
+   "            :class:`Database` instance.\n"
+   "        flags (int): Currently unused.\n"
    "        context (:obj:`object`): A context object passed as the last\n"
    "            arg to **match_event_handler**.\n\n"},
   {NULL}};
@@ -764,9 +773,9 @@ static PyTypeObject DatabaseType = {
   "Database(scratch=None, mode=0)\n\n"
   "    Represents a Hyperscan database.\n\n"
   "    Args:\n"
-  "        scratch (:class:`~.Scratch`, optional): Thread-specific\n"
+  "        scratch (:class:`Scratch`, optional): Thread-specific\n"
   "            scratch space.\n"
-  "        mode (:obj:`int`, optional): One of :const:`HS_MODE_BLOCK`,\n"
+  "        mode (int, optional): One of :const:`HS_MODE_BLOCK`,\n"
   "            :const:`HS_MODE_STREAM`, or :const:`HS_MODE_VECTORED`.\n"
   "        chimera (bool): Enable Chimera support."
   "\n\n",                  /* tp_doc */
@@ -979,8 +988,8 @@ static PyMemberDef Stream_members[] = {
    T_OBJECT_EX,
    offsetof(Stream, database),
    0,
-   ":class:`~.Database`: Database instance."},
-  {"flags", T_INT, offsetof(Stream, flags), 0, ":obj:`int`: Stream flags."},
+   ":class:`Database`: Database instance."},
+  {"flags", T_INT, offsetof(Stream, flags), 0, "int: Stream flags."},
   {NULL}};
 
 static PyMethodDef Stream_methods[] = {
@@ -992,7 +1001,7 @@ static PyMethodDef Stream_methods[] = {
    "close(scratch=None, match_event_handler=None, context=None)\n\n"
    "    Closes the stream.\n\n"
    "    Args:\n"
-   "        scratch (:class:`~.Scratch`, optional): Scratch space.\n"
+   "        scratch (:class:`Scratch`, optional): Scratch space.\n"
    "            If a **match_event_handler** is provided and a scratch\n"
    "            space is not, the scratch space of the associated\n"
    "            database will be used.\n"
@@ -1005,12 +1014,13 @@ static PyMethodDef Stream_methods[] = {
   {"scan",
    (PyCFunction)Stream_scan,
    METH_VARARGS | METH_KEYWORDS,
-   "scan(data, flags=0, scratch=None, match_event_handler, context=None)\n\n"
+   "scan(data, flags=0, scratch=None, match_event_handler=None, "
+   "context=None)\n\n"
    "    Scans streaming text.\n\n"
    "    Args:\n"
-   "        data (:obj:`str`): The block of text to scan.\n"
-   "        flags (:obj:`int`, optional): Currently unused.\n"
-   "        scratch (:obj:`~.Scratch`, optional): Scratch space.\n"
+   "        data (str): The block of text to scan.\n"
+   "        flags (int, optional): Currently unused.\n"
+   "        scratch (:obj:`Scratch`, optional): Scratch space.\n"
    "        match_event_handler (callable, optional): The match \n"
    "            callback, which is invoked for each match result, and\n"
    "            passed the expression id, start offset, end offset,\n"
@@ -1050,9 +1060,9 @@ static PyTypeObject StreamType = {
   "Stream(database=None, flags=0, match_event_handler=None)\n\n"
   "    Provides a context manager for scanning streams of text.\n\n"
   "    Args:\n"
-  "        database (:class:`~.Database`): A database initialized with\n"
+  "        database (:class:`Database`): A database initialized with\n"
   "            :const:`HS_MODE_STREAM`.\n"
-  "        flags (:obj:`int`, optional): Currently unused.\n"
+  "        flags (int, optional): Currently unused.\n"
   "        match_event_handler (callable, optional): The match callback,\n"
   "            which is invoked for each match result, and passed the\n"
   "            expression id, start offset, end offset, flags, and a\n"
@@ -1148,7 +1158,7 @@ static PyMemberDef Scratch_members[] = {
    T_OBJECT_EX,
    offsetof(Scratch, database),
    0,
-   ":class:`~.Database`: The database associated with this scratch space."},
+   ":class:`Database`: The database associated with this scratch space."},
   {NULL}};
 
 static PyMethodDef Scratch_methods[] = {
@@ -1158,7 +1168,7 @@ static PyMethodDef Scratch_methods[] = {
    "clone()\n\n"
    "    Clones a scratch space.\n\n"
    "    Returns:\n"
-   "        :class:`~.Scratch`: The cloned scratch space.\n\n"},
+   "        :class:`Scratch`: The cloned scratch space.\n\n"},
   {"set_database",
    (PyCFunction)Scratch_set_database,
    METH_VARARGS | METH_KEYWORDS,
@@ -1189,7 +1199,7 @@ static PyTypeObject ScratchType = {
   "Scratch(database=None)\n\n"
   "    Represents Hyperscan 'scratch space.'\n\n"
   "    Args:\n"
-  "        database (:class:`~.Database`, optional): A database instance."
+  "        database (:class:`Database`, optional): A database instance."
   "\n\n",                 /* tp_doc */
   0,                      /* tp_traverse */
   0,                      /* tp_clear */
@@ -1214,6 +1224,7 @@ static PyObject *dumpb(PyObject *self, PyObject *args, PyObject *kwds)
   char *buf;
   size_t length;
   static char *kwlist[] = {"database", NULL};
+
   if (!PyArg_ParseTupleAndKeywords(
         args, kwds, "O!", kwlist, &DatabaseType, &db))
     return NULL;
@@ -1229,38 +1240,36 @@ static PyObject *dumpb(PyObject *self, PyObject *args, PyObject *kwds)
     PyErr_SetString(HyperscanError, "failed to serialize database");
     return NULL;
   }
-  Py_INCREF(bytes);
+  PyMem_Free(buf);
   return bytes;
 }
 
 static PyObject *loadb(PyObject *self, PyObject *args, PyObject *kwds)
 {
   char *buf;
-  PyObject *obuf, *ocreatescratch = Py_True;
-  static char *kwlist[] = {"buf", "create_scratch", NULL};
-  if (!PyArg_ParseTupleAndKeywords(
-        args, kwds, "O|O", kwlist, &obuf, &ocreatescratch))
+  PyObject *obuf = Py_None;
+  PyObject *odb;
+  odb = PyObject_CallFunctionObjArgs((PyObject *)&DatabaseType, NULL);
+  Database *db = (Database *)odb;
+
+  static char *kwlist[] = {"buf", "mode", NULL};
+  if (!PyArg_ParseTupleAndKeywords(args, kwds, "OI", kwlist, &obuf, &db->mode))
     return NULL;
   if (!PyBytes_Check(obuf)) {
     PyErr_SetString(PyExc_TypeError, "buf must be a bytestring");
     return NULL;
   }
-  PyObject *odb;
-  odb = PyObject_CallFunctionObjArgs((PyObject *)&DatabaseType, NULL);
-  Database *db = (Database *)odb;
-  Py_INCREF(odb);
+
   Py_ssize_t length = PyBytes_Size(obuf);
   buf = PyBytes_AsString(obuf);
   hs_error_t err = hs_deserialize_database(buf, length, &db->hs_db);
   HANDLE_HYPERSCAN_ERR(err, NULL);
-  if (PyObject_IsTrue(ocreatescratch))
-    db->scratch = PyObject_CallFunction((PyObject *)&ScratchType, "O", odb, 0);
   if (PyErr_Occurred())
     return NULL;
   return odb;
 }
 
-static PyMethodDef Hyperscan_methods[] = {
+static PyMethodDef HyperscanMethods[] = {
   {"dumpb",
    (PyCFunction)dumpb,
    METH_VARARGS | METH_KEYWORDS,
@@ -1273,32 +1282,52 @@ static PyMethodDef Hyperscan_methods[] = {
   {"loadb",
    (PyCFunction)loadb,
    METH_VARARGS | METH_KEYWORDS,
-   "loadb(buf)\n"
+   "loadb(buf, mode)\n"
    "    Deserializes a Hyperscan database.\n\n"
    "    Args:\n"
-   "        buf (:obj:`bytearray`): A serialized Hyperscan database.\n\n"
+   "        buf (:obj:`bytearray`): A serialized Hyperscan database.\n"
+   "        mode (int): The expected mode of the database.\n\n"
    "    Returns:\n"
    "        :class:`Database`: The deserialized database instance.\n\n"},
   {NULL}};
 
-MOD_INIT(_hyperscan)
+static struct PyModuleDef hyperscanmodule = {
+  PyModuleDef_HEAD_INIT,
+  "_ext",
+  "Hyperscan bindings for CPython.",
+  -1,
+  HyperscanMethods,
+};
+
+PyMODINIT_FUNC PyInit__ext(void)
 {
   PyObject *m;
 
-  MOD_DEF(
-    m, "_hyperscan", "Hyperscan bindings for CPython.", Hyperscan_methods);
+  m = PyModule_Create(&hyperscanmodule);
+  if (m == NULL)
+    return NULL;
 
-  if (!m)
-    return MOD_ERROR_VAL;
-
-  ADD_INT_CONSTANT(m, HS_ARCH_ERROR);
-  ADD_INT_CONSTANT(m, HS_BAD_ALIGN);
-  ADD_INT_CONSTANT(m, HS_BAD_ALLOC);
-  ADD_INT_CONSTANT(m, HS_COMPILER_ERROR);
+  ADD_INT_CONSTANT(m, CH_BAD_ALIGN);
+  ADD_INT_CONSTANT(m, CH_BAD_ALLOC);
+  ADD_INT_CONSTANT(m, CH_COMPILER_ERROR);
+  ADD_INT_CONSTANT(m, CH_DB_MODE_ERROR);
+  ADD_INT_CONSTANT(m, CH_DB_PLATFORM_ERROR);
+  ADD_INT_CONSTANT(m, CH_DB_VERSION_ERROR);
+  ADD_INT_CONSTANT(m, CH_FAIL_INTERNAL);
+  ADD_INT_CONSTANT(m, CH_FLAG_CASELESS);
+  ADD_INT_CONSTANT(m, CH_FLAG_DOTALL);
+  ADD_INT_CONSTANT(m, CH_FLAG_MULTILINE);
+  ADD_INT_CONSTANT(m, CH_FLAG_SINGLEMATCH);
+  ADD_INT_CONSTANT(m, CH_FLAG_UCP);
+  ADD_INT_CONSTANT(m, CH_FLAG_UTF8);
+  ADD_INT_CONSTANT(m, CH_INVALID);
+  ADD_INT_CONSTANT(m, CH_MODE_GROUPS);
+  ADD_INT_CONSTANT(m, CH_MODE_NOGROUPS);
+  ADD_INT_CONSTANT(m, CH_NOMEM);
+  ADD_INT_CONSTANT(m, CH_SCAN_TERMINATED);
+  ADD_INT_CONSTANT(m, CH_SCRATCH_IN_USE);
+  ADD_INT_CONSTANT(m, CH_SUCCESS);
   ADD_INT_CONSTANT(m, HS_CPU_FEATURES_AVX2);
-  ADD_INT_CONSTANT(m, HS_DB_MODE_ERROR);
-  ADD_INT_CONSTANT(m, HS_DB_PLATFORM_ERROR);
-  ADD_INT_CONSTANT(m, HS_DB_VERSION_ERROR);
   ADD_INT_CONSTANT(m, HS_EXT_FLAG_EDIT_DISTANCE);
   ADD_INT_CONSTANT(m, HS_EXT_FLAG_HAMMING_DISTANCE);
   ADD_INT_CONSTANT(m, HS_EXT_FLAG_MAX_OFFSET);
@@ -1315,7 +1344,6 @@ MOD_INIT(_hyperscan)
   ADD_INT_CONSTANT(m, HS_FLAG_SOM_LEFTMOST);
   ADD_INT_CONSTANT(m, HS_FLAG_UCP);
   ADD_INT_CONSTANT(m, HS_FLAG_UTF8);
-  ADD_INT_CONSTANT(m, HS_INVALID);
   ADD_INT_CONSTANT(m, HS_MODE_BLOCK);
   ADD_INT_CONSTANT(m, HS_MODE_NOSTREAM);
   ADD_INT_CONSTANT(m, HS_MODE_SOM_HORIZON_LARGE);
@@ -1323,9 +1351,7 @@ MOD_INIT(_hyperscan)
   ADD_INT_CONSTANT(m, HS_MODE_SOM_HORIZON_SMALL);
   ADD_INT_CONSTANT(m, HS_MODE_STREAM);
   ADD_INT_CONSTANT(m, HS_MODE_VECTORED);
-  ADD_INT_CONSTANT(m, HS_NOMEM);
   ADD_INT_CONSTANT(m, HS_OFFSET_PAST_HORIZON);
-  ADD_INT_CONSTANT(m, HS_SCAN_TERMINATED);
   ADD_INT_CONSTANT(m, HS_SCRATCH_IN_USE);
   ADD_INT_CONSTANT(m, HS_SUCCESS);
   ADD_INT_CONSTANT(m, HS_TUNE_FAMILY_BDW);
@@ -1337,47 +1363,158 @@ MOD_INIT(_hyperscan)
   ADD_INT_CONSTANT(m, HS_TUNE_FAMILY_SKX);
   ADD_INT_CONSTANT(m, HS_TUNE_FAMILY_SLM);
   ADD_INT_CONSTANT(m, HS_TUNE_FAMILY_SNB);
-  ADD_INT_CONSTANT(m, CH_MODE_NOGROUPS);
-  ADD_INT_CONSTANT(m, CH_MODE_GROUPS);
-  ADD_INT_CONSTANT(m, CH_FLAG_CASELESS);
-  ADD_INT_CONSTANT(m, CH_FLAG_DOTALL);
-  ADD_INT_CONSTANT(m, CH_FLAG_MULTILINE);
-  ADD_INT_CONSTANT(m, CH_FLAG_SINGLEMATCH);
-  ADD_INT_CONSTANT(m, CH_FLAG_UTF8);
-  ADD_INT_CONSTANT(m, CH_FLAG_UCP);
-  ADD_INT_CONSTANT(m, CH_SUCCESS);
-  ADD_INT_CONSTANT(m, CH_INVALID);
-  ADD_INT_CONSTANT(m, CH_NOMEM);
-  ADD_INT_CONSTANT(m, CH_SCAN_TERMINATED);
-  ADD_INT_CONSTANT(m, CH_COMPILER_ERROR);
-  ADD_INT_CONSTANT(m, CH_DB_VERSION_ERROR);
-  ADD_INT_CONSTANT(m, CH_DB_PLATFORM_ERROR);
-  ADD_INT_CONSTANT(m, CH_DB_MODE_ERROR);
-  ADD_INT_CONSTANT(m, CH_BAD_ALIGN);
-  ADD_INT_CONSTANT(m, CH_BAD_ALLOC);
-  ADD_INT_CONSTANT(m, CH_SCRATCH_IN_USE);
-  ADD_INT_CONSTANT(m, CH_FAIL_INTERNAL);
 
-  HyperscanError = PyErr_NewException("hyperscan.error", NULL, NULL);
-  Py_INCREF(HyperscanError);
-  PyModule_AddObject(m, "error", HyperscanError);
+  HyperscanError = PyErr_NewExceptionWithDoc(
+    "hyperscan.error",
+    "Base exception class for Hyperscan errors.",
+    NULL,
+    NULL);
+  if (!HyperscanError) {
+    goto cleanup_module;
+  }
+  Py_XINCREF(HyperscanError);
+  if (PyModule_AddObject(m, "error", HyperscanError) < 0) {
+    Py_XDECREF(HyperscanError);
+    Py_CLEAR(HyperscanError);
+    goto cleanup_module;
+  }
+  if (PyModule_AddObject(m, "HyperscanError", HyperscanError) < 0) {
+    Py_XDECREF(HyperscanError);
+    Py_CLEAR(HyperscanError);
+    goto cleanup_module;
+  }
+
+  ADD_HYPERSCAN_ERROR(
+    m,
+    HyperscanErrors,
+    HyperscanError,
+    InvalidError,
+    HS_INVALID,
+    "Parameter passed to this function was invalid.");
+  ADD_HYPERSCAN_ERROR(
+    m,
+    HyperscanErrors,
+    HyperscanError,
+    NoMemoryError,
+    HS_NOMEM,
+    "Memory allocation failed.");
+  ADD_HYPERSCAN_ERROR(
+    m,
+    HyperscanErrors,
+    HyperscanError,
+    CompilerError,
+    HS_COMPILER_ERROR,
+    "Pattern compilation failed.");
+  ADD_HYPERSCAN_ERROR(
+    m,
+    HyperscanErrors,
+    HyperscanError,
+    ScanTerminated,
+    HS_SCAN_TERMINATED,
+    "The engine was terminated by callback.");
+  ADD_HYPERSCAN_ERROR(
+    m,
+    HyperscanErrors,
+    HyperscanError,
+    DatabaseVersionError,
+    HS_DB_VERSION_ERROR,
+    "The given database was built for a different version of Hyperscan.");
+  ADD_HYPERSCAN_ERROR(
+    m,
+    HyperscanErrors,
+    HyperscanError,
+    DatabasePlatformError,
+    HS_DB_PLATFORM_ERROR,
+    "The given database was built for a different platform (i.e., CPU type).");
+  ADD_HYPERSCAN_ERROR(
+    m,
+    HyperscanErrors,
+    HyperscanError,
+    DatabaseModeError,
+    HS_DB_MODE_ERROR,
+    "The given database was built for a different mode of operation.");
+  ADD_HYPERSCAN_ERROR(
+    m,
+    HyperscanErrors,
+    HyperscanError,
+    BadAlignError,
+    HS_BAD_ALIGN,
+    "A parameter passed to this function was not correctly aligned.");
+  ADD_HYPERSCAN_ERROR(
+    m,
+    HyperscanErrors,
+    HyperscanError,
+    BadAllocationError,
+    HS_BAD_ALLOC,
+    "The memory allocator failed.");
+  ADD_HYPERSCAN_ERROR(
+    m,
+    HyperscanErrors,
+    HyperscanError,
+    ScratchInUseError,
+    HS_SCRATCH_IN_USE,
+    "The scratch region was already in use.");
+  ADD_HYPERSCAN_ERROR(
+    m,
+    HyperscanErrors,
+    HyperscanError,
+    ArchitectureError,
+    HS_ARCH_ERROR,
+    "Unsupported CPU architecture.");
+  ADD_HYPERSCAN_ERROR(
+    m,
+    HyperscanErrors,
+    HyperscanError,
+    InsufficientSpaceError,
+    HS_INSUFFICIENT_SPACE,
+    "Provided buffer was too small.");
+  ADD_HYPERSCAN_ERROR(
+    m,
+    HyperscanErrors,
+    HyperscanError,
+    UnknownError,
+    HS_UNKNOWN_ERROR,
+    "Unexpected internal error.");
+  ADD_HYPERSCAN_ERROR(
+    m,
+    HyperscanErrors,
+    HyperscanError,
+    InternalPCREError,
+    CH_FAIL_INTERNAL,
+    "Unexpected internal error.");
 
   if (
-    PyType_Ready(&DatabaseType) < 0 || PyType_Ready(&ScratchType) < 0 ||
-    PyType_Ready(&StreamType) < 0)
-    return MOD_ERROR_VAL;
+    (PyType_Ready(&DatabaseType) < 0) || (PyType_Ready(&ScratchType) < 0) ||
+    (PyType_Ready(&StreamType) < 0)) {
+    goto cleanup_module;
+  }
 
-  Py_INCREF(&DatabaseType);
-  PyModule_AddObject(m, "Database", (PyObject *)&DatabaseType);
+  Py_XINCREF(&DatabaseType);
+  if (PyModule_AddObject(m, "Database", (PyObject *)&DatabaseType) < 0) {
+    Py_XDECREF(&DatabaseType);
+    goto cleanup_module;
+  }
 
   ScratchType.tp_new = PyType_GenericNew;
-  Py_INCREF(&ScratchType);
-  PyModule_AddObject(m, "Scratch", (PyObject *)&ScratchType);
+  Py_XINCREF(&ScratchType);
+  if (PyModule_AddObject(m, "Scratch", (PyObject *)&ScratchType) < 0) {
+    Py_XDECREF(&ScratchType);
+    goto cleanup_module;
+  }
 
-  Py_INCREF(&StreamType);
-  PyModule_AddObject(m, "Stream", (PyObject *)&StreamType);
+  Py_XINCREF(&StreamType);
+  if (PyModule_AddObject(m, "Stream", (PyObject *)&StreamType) < 0) {
+    Py_XDECREF(&StreamType);
+    goto cleanup_module;
+  }
 
-  PyModule_AddStringConstant(m, "__version__", hs_version());
+  if (PyModule_AddStringConstant(m, "__version__", hs_version()) < 0) {
+    goto cleanup_module;
+  }
 
-  return MOD_SUCCESS_VAL(m);
+  return m;
+
+cleanup_module:
+  Py_DECREF(m);
+  return NULL;
 }
